// ********************************************************************
// *** BT 747 ***
// *** April 14, 2007 ***
// *** (c)2007 Mario De Weerd ***
// *** m.deweerd@ieee.org ***
// *** ********************************************************** ***
// *** Software is provided "AS IS," without a warranty of any ***
// *** kind. ALL EXPRESS OR IMPLIED REPRESENTATIONS AND WARRANTIES,***
// *** INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS ***
// *** FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY ***
// *** EXCLUDED. THE ENTIRE RISK ARISING OUT OF USING THE SOFTWARE ***
// *** IS ASSUMED BY THE USER. ***
// *** See the GNU General Public License Version 3 for details. ***
// *** *********************************************************** ***
package gps.connection;

import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.UnsupportedCommOperationException;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;

import bt747.Version;
import bt747.sys.Generic;

/**
 * This class implements the serial port for rxtx (on Linux).
 * 
 * @author Mario De Weerd
 */
public final class GPSRxTxPort extends GPSPort {
    private CommPort sp = null;

    private OutputStream ds = null;
    private InputStream in = null;
    private String portPrefix = "";
    private boolean hasPortNbr = true;
    public static final String os_name = java.lang.System
            .getProperty("os.name");
    public static boolean dubiousTried = false;
  
    /**
     * 
     */
    public GPSRxTxPort() {
        super();

        if (os_name.startsWith("Windows")) {
            portPrefix = "COM";
            dubiousTried = true; // Dubious not needed on Win
        } else if (os_name.startsWith("Linux")) {
            portPrefix = "/dev/ttyUSB";
        } else if (os_name.startsWith("Mac")) {
            dubiousTried = true; // Dubious not needed on Mac
            setUSB();
            setBlueTooth();
            // portPrefix="/dev/tty.iBt-GPS-SPPslave-";
            portPrefix = "/dev/cu.SLAB_USBtoUART";
            hasPortNbr = false;
        }
        portPrefix = java.lang.System.getProperty("bt747_prefix", portPrefix);
    }

    /**
     * Indicates if the device is connected or not.
     * 
     * @return <code>true</code> if the device is connected.
     */
    public final synchronized boolean isConnected() {
        return (ds != null);
    }

    /**
     * Close the connection.
     * 
     * 
     */
    public final synchronized void closePort() {
        if (ds != null) {
            try {
                in.close();
                in = null;
            } catch (final Exception e) {
                Generic.debug("", e);
            }
            try {
                ds.close();
                ds = null;
            } catch (final Exception e) {
                Generic.debug("", e);
            }
            try {
                if (sp != null) {
                    sp.close();
                    sp = null;
                }
            } catch (final Exception e) {
                Generic.debug("", e);
            }
        }
    }

    private final void openDubiousPortAsWorkaroundOnLinux() {
        if (!dubiousTried) {
            dubiousTried = true;
            try {
                CommPortIdentifier portIdentifier;
                portIdentifier = CommPortIdentifier
                        .getPortIdentifier("/dev/dubiousport");

                final CommPort commPort = portIdentifier.open(
                        "BT747 " + Version.VERSION_NUMBER + " "
                                + getClass().getName(), 2000);
                commPort.close();
            } catch (Throwable e) {
                // Do nothing, this is expected.
            }

        }
    }
    /**
     * Open a connection.
     * 
     * @return status result of the opening of the serial port.
     */
    public final synchronized int openPort() {
        int result = -1;
        String portStr;
        if (getFreeTextPort().length() == 0) {
            portStr = java.lang.System.getProperty("bt747_port", portPrefix
                    + (hasPortNbr ? spPortNbr : ""));
        } else {
            portStr = getFreeTextPort();
        }

        closePort();

        openDubiousPortAsWorkaroundOnLinux();
        
        try {
            if (Generic.isDebug()) {
                Generic.debug("Info: trying to open '" + portStr + "'", null);
            }
            CommPortIdentifier portIdentifier;
            portIdentifier = CommPortIdentifier.getPortIdentifier(portStr);

                final CommPort commPort = portIdentifier.open(
                        "BT747 " + Version.VERSION_NUMBER + " "
                                + getClass().getName(), 2000);
                if (Generic.isDebug()) {
                    Generic.debug("Info: Opened port, setting parameters", null);
                }
                if (commPort instanceof SerialPort) {
                    final SerialPort serialPort = (SerialPort) commPort;
                    sp = serialPort;
                    in = sp.getInputStream();
                    ds = sp.getOutputStream();
                    final int baud = getSpeed();
                    final int databits = SerialPort.DATABITS_8;
                    int stopbits = SerialPort.STOPBITS_1;
                    final int parity = SerialPort.PARITY_NONE;
                    boolean parametersSet = false;
                    try {
                        serialPort.setSerialPortParams(baud, databits,
                                stopbits, parity);
                        parametersSet = true;
                    } catch (Throwable e) {
                        Generic.debug("Issue when setting parameters:" + baud
                                + " " + databits + " " + stopbits + " "
                                + parity, e);
                    }
                    if(!parametersSet) {
                        stopbits = SerialPort.STOPBITS_2;
                        Generic.debug("Trying stopbits = "+stopbits, null);
                        try {
                            serialPort.setSerialPortParams(baud, databits,
                                    stopbits, parity);
                            parametersSet = true;
                        } catch (Throwable e) {
                            Generic.debug("Issue when setting parameters:" + baud
                                    + " " + databits + " " + stopbits + " "
                                    + parity, e);
                        }
                    }
                    result = 0;
                } else {
                    sp = null;
                    in = null;
                    ds = null;
                    System.out
                            .println("Error: Only serial ports are handled.");
                }
        } catch (final NoSuchPortException e) {
            Generic.debug("", e);
            Generic.debug("\nListing known ports:");
            try {
                final Enumeration<?> list = CommPortIdentifier
                        .getPortIdentifiers();
                while (list.hasMoreElements()) {
                    final CommPortIdentifier iden = (CommPortIdentifier) list
                            .nextElement();
                    String type;
                    switch (iden.getPortType()) {
                    case CommPortIdentifier.PORT_SERIAL: // rs232 Port
                        type = "SER  :";
                        break;
                    case CommPortIdentifier.PORT_PARALLEL: // Parallel Port
                        type = "PAR  :";
                        break;
                    case CommPortIdentifier.PORT_I2C: // i2c Port
                        type = "I2C  :";
                        break;
                    case CommPortIdentifier.PORT_RS485: // rs485 Port
                        type = "RS485:";
                        break;
                    case CommPortIdentifier.PORT_RAW: // Raw
                        // Port
                        type = "RAW  :";
                        break;
                    default:
                        type = "UNK  :";
                    }
                    Generic.debug(type + iden.getName());
                }
            } catch (Exception exp) {
                // Do not care.

            }
        } catch (final PortInUseException e) {
            Generic.debug("Error: Port " + portStr
                    + "is currently in use", e);
//        } catch (final UnsupportedCommOperationException e) {
//            Generic.debug("", e);
        } catch (final IOException e) {
            Generic.debug("", e);
        }

        return result;
    }

    private final boolean isValidPort(final String path) {
        try {
            return (new File(path)).exists()
                    && (CommPortIdentifier.getPortIdentifier(path)
                            .getPortType() == CommPortIdentifier.PORT_SERIAL);
        } catch (final Exception e) {
            return false;
        }

    }

    /** Maximum index when looking for BT port number given a base name. */
    private static final int MAX_BTPORT_SEARCH_IDX = 3;
    /** Maximum index when looking for USB port number given a base name. */
    private static final int MAX_USBPORT_SEARCH_IDX = 6;

    /**
     * List of valid bluetooth port prefixes for the mac.
     */
    private static final String[] macPortPrefixes = {
            "/dev/tty.HOLUX_M-241-SPPSlave-", // Port for Holux M-241
            "/dev/tty.HoluxM-1000C-SPPslave-", // Port for Holux M1000C
            "/dev/tty.iBT-GPS-SPPSlave-", // Port for
            "/dev/tty.iBT-GPS-SPPslave-", // Port for
            "/dev/cu.QstarzGPS-SPPslave-", // Port for some Qstartz devices
            "/dev/tty.QstarzGPS-SPPslave-",// Port for some Qstartz devices
            "/dev/tty.BlumaxBT-GPS-SPPSlave-" // Port for Blumax device
    };

    /**
     * Set a bluetooth connection.
     */
    public final synchronized void setBlueTooth() {
        super.setBlueTooth();
        spPortNbr = 0;

        boolean portFound = false;
        if (os_name.toLowerCase().startsWith("mac")) {
            portFound = setPortFromPrefixes(macPortPrefixes,
                    MAX_BTPORT_SEARCH_IDX);
        }

        if (!portFound) {
            setFreeTextPort("");
        }
    }

    /**
     * Goes through a list of port prefixes for indexes from 0 up to maxIdx.
     * 
     * @param prefixes
     *            list of prefixes for ports that will be appended with an
     *            index.
     * @param maxIdx
     *            maximum index for ports.
     * @return true if a valid port was found and set.
     */
    private final synchronized boolean setPortFromPrefixes(
            final String[] prefixes, final int maxIdx) {
        boolean portFound = false;
        for (int prefixIdx = 0; !portFound && prefixIdx < prefixes.length; prefixIdx++) {
            for (int i = 0; !portFound && (i < maxIdx); i++) {
                // /dev/tty.HOLUX_M-241-SPPSlave-1
                setFreeTextPort(prefixes[prefixIdx] + i);
                portFound = isValidPort(getFreeTextPort());
            }
        }
        return portFound;
    }

    private static final String[] linUsbPortPrefixes = { //
    "/dev/ttyUSB", // Prefix on linux for USB ports
            "/dev/ttyACM" // Prefix on linux for newer USB ports of GPS
    };

    /**
     * Set an USB connection.
     */
    public final synchronized void setUSB() {
        super.setUSB();
        // POSSIBLE_PORTS="$POSSIBLE_PORTS /dev/cu.SLAB_USBtoUART"
        // POSSIBLE_PORTS="$POSSIBLE_PORTS /dev/tty.HOLUX_M-241-SPPSlave-1"
        // POSSIBLE_PORTS="$POSSIBLE_PORTS /dev/tty.HOLUX_M-241-SPPSlave-0"
        // POSSIBLE_PORTS="$POSSIBLE_PORTS /dev/tty.iBT-GPS-SPPSlave-1"
        // POSSIBLE_PORTS="$POSSIBLE_PORTS /dev/cu.serial-0001"
        // POSSIBLE_PORTS="$POSSIBLE_PORTS /dev/tty.iBT-GPS-SPPslave-1"
        boolean portFound = false;
        if (!portFound) {
            setFreeTextPort("/dev/cu.serial-0001");
            portFound = isValidPort(getFreeTextPort());
        }
        if (!portFound) {
            setFreeTextPort("/dev/cu.SLAB_USBtoUART");
            portFound = isValidPort(getFreeTextPort());
        }
        if (!portFound) {
            setFreeTextPort("/dev/cu.usbmodem1b10");
            portFound = isValidPort(getFreeTextPort());
        }
        if (!portFound) {
            setFreeTextPort("/dev/cu.usbmodem1d10");
            portFound = isValidPort(getFreeTextPort());
        }
        if (!portFound) {
            setFreeTextPort("/dev/cu.usbmodem3a20");
            portFound = isValidPort(getFreeTextPort());
        }
        if (!portFound) {
            setFreeTextPort("/dev/cu.iBTAGPS-SPPslave");
            portFound = isValidPort(getFreeTextPort());
        }

        if (!portFound && os_name.toLowerCase().startsWith("lin")) {
            portFound = setPortFromPrefixes(linUsbPortPrefixes,
                    MAX_USBPORT_SEARCH_IDX);
        }
        if (!portFound) {
            setFreeTextPort("");
        }
    }

    public final synchronized void setFreeTextPort(final String s) {
        super.setFreeTextPort(s);
    }

    public final synchronized String getFreeTextPort() {
        return super.getFreeTextPort();
    }

    /**
     * getter to retrieve the last error report by the serial port driver.
     * 
     * @return last error from the SerialPort driver
     */
    public final int error() {
        return 0;
    }

    public final synchronized void write(final byte[] b) {
        try {
            if (ds != null) {
                ds.write(b);
            }
        } catch (final Exception e) {
            Generic.debug("", e);
        }
        if (GPS_FILE_LOG && (debugFile != null)) {
            try {
                debugFile.writeBytes("\nWrite:".getBytes(), 0, 2);
                debugFile.writeBytes(b, 0, b.length);
            } catch (final Exception e) {
                Generic.debug("", e);
            }
        }
    }

    public final void write(final String s) {
        write(s.getBytes());
    }

    public final synchronized int readCheck() {
        if (in != null) {
            try {
                // System.err.println("Available: "+in.available());
                // return 100;
                return in.available();
            } catch (final Exception e) {
                Generic.debug("", e);
                return 0;
            }
        } else {
            return 0;
        }
    }

    public final synchronized int readBytes(final byte[] b, final int start,
            final int max) {
        try {
            return in.read(b, start, max);
        } catch (final Exception e) {
            Generic.debug("", e);
            return 0;
        }
    }
}
